/**
 *	@file 	database.c
 *	@brief	A tap database for recognized files and prg data.
 *
 *	Details here.
 */
#ifdef WIN32
#include <windows.h>
#include <direct.h>
#endif
#include "mydefs.h"
#include "database.h"
#include "main.h"

struct blk_t *blk[BLKMAX];	/*!< Database of all found entities. */
struct prg_t prg[BLKMAX];	/*!< Database of all extracted files (prg's) */
int dbase_is_full = FALSE;	/*!< Flag that indicates database capacity
				     reached */
/**
 *	Allocate ram to file database and initialize array pointers.
 *
 *	@param void
 *
 *	@return TRUE on success
 *	@return FALSE on memory allocation failure
 */

int create_database(void)
{
	int i;

	for (i = 0; i < BLKMAX; i++) {
		blk[i] = (struct blk_t*)malloc(sizeof(struct blk_t));
		if (blk[i] == NULL) {
			printf("\nError: malloc failure whilst creating file database.");
			destroy_database(); /* Free any already allocated resource */
			return FALSE;
		}

		blk[i]->dd = NULL;
		blk[i]->fn = NULL;
	}

	return TRUE;
}

/**
 *	Clear database
 *
 *	@param void
 *
 *	@return none
 */

void reset_database(void)
{
	int i;

	for (i = 0 ; i < BLKMAX; i++) {		/*!< clear database... */
		blk[i]->lt = 0;
		blk[i]->p1 = 0;
		blk[i]->p2 = 0;
		blk[i]->p3 = 0;
		blk[i]->p4 = 0;
		blk[i]->xi = 0;
		blk[i]->cs = 0;
		blk[i]->ce = 0;
		blk[i]->cx = 0;
		blk[i]->crc = 0;
		blk[i]->rd_err = 0;
		blk[i]->cs_exp = HASNOTCHECKSUM;
		blk[i]->cs_act = HASNOTCHECKSUM;
		blk[i]->pilot_len = 0;
		blk[i]->trail_len = 0;
		blk[i]->ok = 0;

		if (blk[i]->dd != NULL) {
			free(blk[i]->dd);
			blk[i]->dd = NULL;
		}

		if(blk[i]->fn != NULL) {
			free(blk[i]->fn);
			blk[i]->fn = NULL;
		}
	}
}

/**
 *	Add a block definition (file details) to the database (blk)
 *
 *	Only sof & eof must be assigned (legal) values for the block,
 *	the others can be 0.
 *
 *	@param lt loader id
 *	@param sof start of block offset
 *	@param sod start of data offset
 *	@param eod end of data offset
 *	@param eof end of block offset
 *	@param xi extra info offset
 *
 *	@return Slot number the block went to
 *	@return DBERR on invalid block definition
 *	@return DBFULL on database full
 */

int addblockdef(int lt, int sof, int sod, int eod, int eof, int xi)
{
	int i, slot, e1, e2;

	if (debug == FALSE) {

		/* check that the block does not conflict with any existing blocks... */

		for (i = 0; blk[i]->lt != 0; i++) {
			e1 = blk[i]->p1;	/* get existing block start pos  */
			e2 = blk[i]->p4;	/* get existing block end pos   */

			if (!((sof < e1 && eof < e1) || (sof > e2 && eof > e2)))
			{
				if(blk[i]->lt==2) /* iAN: if there's a pause between data stream, who cares??*/
					blk[i]->lt=0;
				else
					return DBERR;
			}
		}
	}

	if ((sof > 19 && eof < tap.len) && (eof >= sof)) {

		/* find the first free slot (containing 0 in 'lt' field)... */

		/* note: slot blk[BLKMAX-1] is reserved for the list terminator. */
		/* the last usable slot is therefore BLKMAX-2. */

		for (i = 0; blk[i]->lt != 0; i++);
			slot = i;

		if (slot == BLKMAX-1) {	/* only clear slot is the last one? (the terminator) */
			if (dbase_is_full == FALSE) {	/* we only need give the error once */
				if (!batchmode)		/* dont bother with the warning in batch mode.. */
					msgout("\n\nWarning: FT's database is full...\nthe report will not be complete.\nTry optimizing.\n\n");
				dbase_is_full = TRUE;
			}
			return DBFULL;
		} else {

			/* put the block in the last available slot... */

			blk[slot]->lt = lt;
			blk[slot]->p1 = sof;
			blk[slot]->p2 = sod;
			blk[slot]->p3 = eod;
			blk[slot]->p4 = eof;
			blk[slot]->xi = xi;

			/* just clear out the remaining fields... */

			blk[slot]->cs = 0;
			blk[slot]->ce = 0;
			blk[slot]->cx = 0;
			blk[slot]->dd = NULL;
			blk[slot]->crc = 0;
			blk[slot]->rd_err = 0;
			blk[slot]->cs_exp = HASNOTCHECKSUM;
			blk[slot]->cs_act = HASNOTCHECKSUM;
			blk[slot]->pilot_len = 0;
			blk[slot]->trail_len = 0;
			blk[slot]->fn = NULL;
			blk[slot]->ok = 0;
		}
	} else
		return DBERR;

	return slot;	/* ok, entry added successfully.   */
}

/**
 *	Sort the database by p1 (file start position, sof) values.
 *
 *	@param void
 *
 *	@return none
 */

void sort_blocks(void)
{
	int i, swaps,size;
	struct blk_t *tmp;

	for (i = 0; blk[i]->lt != 0 && i < BLKMAX; i++);

	size = i;	/* store current size of database */

	do {
		swaps = 0;
		for (i = 0; i < size - 1; i++) {

			/* examine file sof's (p1's), swap if necessary... */

			if ((blk[i]->p1) > (blk[i + 1]->p1)) {
				tmp = blk[i];
				blk[i] = blk[i + 1];
				blk[i + 1] = tmp;
				swaps++;
			}
		}
	} while(swaps != 0);	/* repeat til no swaps occur.  */
}

/*
 * Searches the file database for gaps. adds a definition for any found.
 * Note: Must be called ONLY after sorting the database!.
 * Note : The database MUST be re-sorted after a GAP is added!.
 */

void scan_gaps(void)
{
	int i, p1, p2, sz;

	p1 = 20;			/* choose start of TAP and 1st blocks first pulse  */
	p2 = blk[0]->p1;

	if (p1 < p2) {
		sz = p2 - p1;
		addblockdef(GAP, p1, 0, 0, p2 - 1, sz);
		sort_blocks();
	}

	/* double dragon sticks in this loop */

	for (i = 0; blk[i]->lt != 0 && blk[i + 1]->lt != 0; i++) {
		p1 = blk[i]->p4;		/* get end of this block */
		p2 = blk[i + 1]->p1;		/* and start of next  */
		if (p1 < (p2 - 1)) {
			sz = (p2 - 1) - p1;
			if (sz > 0) {
				addblockdef(GAP, p1 + 1, 0, 0, p2 - 1, sz);
				sort_blocks();
			}
		}
	}

	p1 = blk[i]->p4;		/* choose last blocks last pulse and End of TAP */
	p2 = tap.len - 1;
	if (p1 < p2) {
		sz = p2 - p1;
		addblockdef(GAP, p1 + 1, 0, 0, p2, sz);
		sort_blocks();
	}
}

/*
 * Counts the number of standard CBM boot sequence(s) (HEAD,HEAD,DATA,DATA) in
 * the current file database.
 * Returns the number found.
 */

int count_bootparts(void)
{
	int tblk[BLKMAX];
	int i, j, bootparts;

	/* make a block list (types only) without gaps/pauses included.. */

	for (i = 0,j = 0; blk[i]->lt != 0; i++) {
		if (blk[i]->lt > 2)
			tblk[j++] = blk[i]->lt;
	}

	tblk[j] = 0;

	/* count bootable sections... */

	bootparts =0;

	for (i = 0; i + 3 < j; i++) {
		if (tblk[i + 0] == CBM_HEAD && tblk[i + 1] == CBM_HEAD && tblk[i + 2] == CBM_DATA && tblk[i + 3] == CBM_DATA)
			bootparts++;
	}

	return bootparts;
}

/*
 * Returns the number of imperfect pulsewidths found in the file at database
 * entry 'slot'.
 */

int count_unopt_pulses(int slot)
{
	int i, t, b, imperfect;

	t = blk[slot]->lt;
	if(t==ACTIONREPLAY)
	{
		t=ACTIONREPLAY_SUPER;
		ft[t].mp=ft[ACTIONREPLAY].lp;
	}
	for(imperfect = 0, i = blk[slot]->p1; i < blk[slot]->p4 + 1; i++) {
		b = tap.tmem[i];
		if (b != ft[t].sp && b != ft[t].mp && b != ft[t].lp)
			imperfect++;
	}

	return imperfect;
}

/*
 * Return a count of files in the database that are 100% optimized...
 */

int count_opt_files(void)
{
	int i, n;

	for (n = 0, i = 0; blk[i]->lt != 0; i++) {
		if (blk[i]->lt > PAUSE) {
			if (count_unopt_pulses(i) == 0)
				n++;
		}
	}

	return n;
}

/*
 * Counts pauses within the tap file, note: v1 pauses are still held
 * as separate entities in the database even when they run consecutively.
 * This function counts consecutive v1's as a single pause.
 */

int count_pauses(void)
{
	int i, n;

	for (n = 0, i = 0; blk[i]->lt != 0; i++) {
		if (blk[i]->lt == PAUSE) {
			n++;
			if (tap.version == 1) {		/* consecutive v1 pauses count as 1 pause.. */
				while (blk[i++]->lt == PAUSE && i < BLKMAX - 1);
				i--;
			}
		}
	}

	return n;
}

/*
 * Return the number of pulses accounted for in total across all known files.
 */

int count_rpulses(void)
{
	int i, tot;

	/* add up number of pulses accounted for... */

	/* for each block entry in blk */

	for (i = 0, tot = 0; blk[i]->lt != 0; i++) {
		if (blk[i]->lt != GAP) {

			/* start and end addresses both present?  */

			if (blk[i]->p1 != 0 && blk[i]->p4 != 0)
				tot += (blk[i]->p4 - blk[i]->p1) + 1;
		}
	}

	return tot;
}

/*
 * Returns the quantity of 'has checksum and its OK' files in the database.
 */

int count_good_checksums(void)
{
	int i, c;

	for (i = 0,c = 0; blk[i]->lt != 0; i++) {
		if (blk[i]->cs_exp != HASNOTCHECKSUM) {
			if (blk[i]->cs_exp == blk[i]->cs_act)
				c++;
		}
	}

	return c;
}

/*
 * Add together all data file CRC32's
 */

int compute_overall_crc(void)
{
	int i, tot = 0;

	for (i = 0; blk[i]->lt != 0; i++)
		tot += blk[i]->crc;

	return tot;
}

/**
 *	Deallocate file database from RAM
 *
 *	A check for non-NULL is done, in case we are freeing resources
 *	after a malloc failure in create_database() (clean job).
 *
 *	@param void
 *
 *	@return none
 */

void destroy_database(void)
{
	int i;

	for (i = 0; i < BLKMAX && blk[i] != NULL; i++)
		free(blk[i]);
}



/* Still unverified PRG functions*/

/*
 * Create a table of exportable PRGs in the prg[] array based on the current
 * data extractions available in the blk[] array.
 * Note: if 'joinprgneighbours' is not 0, then any neigbouring files will be
 * connected as a single PRG. (neighbour = data addresses run consecutively).
 */

void make_prgs(void)
{
	int i, c, j, t, x, s, e, errors, ti, pt[BLKMAX];
	unsigned char *tmp,*tmpfn=NULL, done;

	/* create table of all exported files by index (of blk)... */

	for (i = 0,j = 0; blk[i]->lt != 0; i++) {
		if (blk[i]->dd != NULL)
			pt[j++] = i;
	}
	pt[j] = -1;			/* terminator */

	/* Clear the prg table... */
	reset_prg_database();

	tmp = (unsigned char*)malloc(65536 * 2);	/* create buffer for unifications */
	j = 0;						/* j steps through the finished prg's */

	/* scan through the 'data holding' blk[] indexes held in pt[]... */

	for (i = 0; pt[i] != -1 ;i++) {
		if (blk[pt[i]]->dd != NULL) {	/* should always be true. */
			ti = 0;
			done = 0;

			s = blk[pt[i]]->cs;	/* keep 1st start address. */
			errors = 0;		/* this will count the errors found in each blk */
						/* entry used to create the final data prg in tmp. */
			do {
				t = blk[pt[i]]->lt;	/* get details of next exportable part... */
							/* note: where files are united, the type will */
							/* be set to the type of only the last file */
				x = blk[pt[i]]->cx;
				e = blk[pt[i]]->ce;
				errors += blk[pt[i]]->rd_err;

				/* bad checksum also counts as an error */

				if (ft[blk[pt[i]]->lt].has_cs == TRUE && blk[pt[i]]->cs_exp != blk[pt[i]]->cs_act)
					errors++;

				for (c = 0; c < x; c++)
					tmp[ti++] = blk[pt[i]]->dd[c];	/* copy the data to tmp buffer */

				/* block unification wanted?... */

				if (prgunite) {

					/* scan following blocks and override default details. */

					if (pt[i + 1] != -1) {		/* another file available? */
						if (blk[pt[i + 1]]->cs == (e + 1))	/* is next file a neighbour? */
							i++;
						else
							done = 1;
					} else
						done =1;
				} else
					done =1;
			} while(!done);

			/* create the finished prg entry using data in tmp...  */

			x = ti;					/* set final data length */
			prg[j].dd = (unsigned char*)malloc(x);	/* allocate the ram */
			for (c = 0; c < x; c++)			/* copy the data.. */
				prg[j].dd[c] = tmp[c];
			prg[j].lt = t;				/* set file type */
			prg[j].cs = s;				/* set file start address */
			prg[j].ce = e;				/* set file end address */
			prg[j].cx = x;				/* set file length */
			prg[j].errors = errors;			/* set file errors */
			if(blk[pt[i]]->fn) /* hinerit last valid fn also in bloack with no fn (cbm header name will be also in turbo header) */
				tmpfn=blk[pt[i]]->fn;
			prg[j].fn     = tmpfn;	/* set file name */

			j++;					/* onto the next prg... */
		}
	}

	prg[j].lt = 0;		/* terminator */

	free(tmp);
}

#ifdef WIN32
int DelAll(char *cmask)
{
    WIN32_FIND_DATA FindFileData;
    HANDLE hFind;
    char temp[MAXPATH]=".\\";
    char *p;
    int i=0;

    p=strrchr(cmask,'\\');
    if(p)
    {
        lstrcpy(temp,cmask);
    }

    p=strrchr(temp,'\\');
    p++;


    hFind = FindFirstFile(cmask, &FindFileData);

    if (hFind != INVALID_HANDLE_VALUE)
    {
        do
        {
            *p='\0';
            lstrcat(temp, FindFileData.cFileName);
            if(DeleteFile(temp))
                ++i;
        } while( FindNextFile(hFind,&FindFileData));

        FindClose(hFind);
    }
    return i;

}
#endif

/*
 * Save all available prg's to a folder (console app only).
 * Returns the number of files saved.
 */

int save_prgs(void)
{
	int i,j;
    unsigned char tmpfn[32];
	FILE *fp;

    if(useprgdir>0)
    {
		char *p;
    	strcpy(prgdir,tap.name);
		p=strrchr(prgdir,'.'); if(p) *p=0;
    	strcat(prgdir,"_prg");
    }

	chdir(exedir);

	if (chdir(prgdir) == 0) {	/* delete old prg's and prg folder if exists... */
#ifdef WIN32
		DelAll("*.prg");
#else
		system(OSAPI_DELETE_FILE" *.prg");
#endif
	} else {
#ifdef WIN32
		CreateDirectory(prgdir,NULL);
#else
		system(OSAPI_CREATE_FOLDER" prg");
#endif
		chdir(prgdir);
	}

	for (i = 0; prg[i].lt != 0; i++) {

        strcpy(tmpfn,prg[i].fn);
        for(j=0;tmpfn[j];j++)
        {
            if( strchr("*<>?:|^",tmpfn[j])!=NULL )
            {
                tmpfn[j]='_';
            }
            else if(strchr(",\\/",tmpfn[j])!=NULL )
            {
                tmpfn[j]='.';
            }
            else if(tmpfn[j] == '\"')
            {
                tmpfn[j]='\'';
            }
            else if(tmpfn[j]>0x7f)
            {
                tmpfn[j]='_';
            }
        }

		sprintf(lin, "%03d %s (%04X-%04X)", i + 1, tmpfn, prg[i].cs, prg[i].ce); /* iAN CooG: include header name */
		if (prg[i].errors != 0)		/* append error indicator if necessary) */
			strcat(lin, " BAD");
		strcat(lin, ".prg");

		fp = fopen(lin, "w+b");
		if (fp != NULL) {
			fputc(prg[i].cs & 0xFF, fp);		/* write low byte of start address */
			fputc((prg[i].cs & 0xFF00) >> 8, fp);	/* write high byte of start address */
			fwrite(prg[i].dd, prg[i].cx, 1, fp);	/* write data */
			fclose(fp);
		}
	}

	chdir(exedir);
	return 0;
}

/* empty the prg datbase...  */
void reset_prg_database(void)
{
	int i;

	for (i = 0; i < BLKMAX; i++) {		/* empty the prg datbase...  */
		prg[i].lt = 0;
		prg[i].cs = 0;
		prg[i].ce = 0;
		prg[i].cx = 0;
		if (prg[i].dd != NULL) {
			free(prg[i].dd);
			prg[i].dd = NULL;
		}

		prg[i].errors = 0;
	}
}
